/*
 * Copyright (c) 2024 NVIDIA CORPORATION & AFFILIATES, ALL RIGHTS RESERVED.
 *
 * This software product is a proprietary product of NVIDIA CORPORATION &
 * AFFILIATES (the "Company") and all right, title, and interest in and to the
 * software product, including all associated intellectual property rights, are
 * and shall remain exclusively with the Company.
 *
 * This software product is governed by the End User License Agreement
 * provided with the software product.
 *
 */

#include <string.h>
#include <unistd.h>

#include <rte_ethdev.h>

#include <doca_log.h>
#include <doca_flow.h>

#include "flow_common.h"

DOCA_LOG_REGISTER(FLOW_RSS_ESP);

/* The number of packets in the rx queue */
#define PACKET_BURST 256

/* The maximal number of queues in RSS forward */
#define MAX_RSS_QUEUES 8

/* Get the percentage according to part and total */
#define GET_PERCENTAGE(part, total) (((double)part / (double)total) * 100)

/* The number of seconds app waits for traffic to come */
#define WAITING_TIME 15

/*
 * Create DOCA Flow pipe with header existence match, copy action, counter and forward RSS.
 *
 * @port [in]: port of the pipe.
 * @nb_rss_queues [in]: number of RSS queues used by pipe FWD action.
 * @pipe [out]: created pipe pointer.
 * @return: DOCA_SUCCESS on success and DOCA_ERROR otherwise.
 */
static doca_error_t create_rss_esp_pipe(struct doca_flow_port *port,
					uint32_t nb_rss_queues,
					struct doca_flow_pipe **pipe)
{
	struct doca_flow_match match;
	struct doca_flow_action_desc desc;
	struct doca_flow_action_descs descs = {.nb_action_desc = 1, .desc_array = &desc};
	struct doca_flow_action_descs *desc_list[] = {&descs};
	struct doca_flow_monitor counter;
	struct doca_flow_fwd fwd;
	struct doca_flow_fwd fwd_miss;
	struct doca_flow_pipe_cfg *pipe_cfg;
	uint16_t rss_queues[MAX_RSS_QUEUES];
	doca_error_t result;
	uint32_t i;

	memset(&match, 0, sizeof(match));
	memset(&desc, 0, sizeof(desc));
	memset(&counter, 0, sizeof(counter));
	memset(&fwd, 0, sizeof(fwd));
	memset(&fwd_miss, 0, sizeof(fwd_miss));

	result = doca_flow_pipe_cfg_create(&pipe_cfg, port);
	if (result != DOCA_SUCCESS) {
		DOCA_LOG_ERR("Failed to create doca_flow_pipe_cfg: %s", doca_error_get_descr(result));
		return result;
	}

	result = set_flow_pipe_cfg(pipe_cfg, "RSS_ESP_PIPE", DOCA_FLOW_PIPE_BASIC, true);
	if (result != DOCA_SUCCESS) {
		DOCA_LOG_ERR("Failed to set doca_flow_pipe_cfg: %s", doca_error_get_descr(result));
		goto destroy_pipe_cfg;
	}

	result = doca_flow_pipe_cfg_set_nr_entries(pipe_cfg, 1);
	if (result != DOCA_SUCCESS) {
		DOCA_LOG_ERR("Failed to set doca_flow_pipe_cfg nr_entries: %s", doca_error_get_descr(result));
		goto destroy_pipe_cfg;
	}

	/* IPv4 and ESP existing match */
	match.parser_meta.outer_l4_type = DOCA_FLOW_L4_META_ESP;
	match.parser_meta.outer_l3_type = DOCA_FLOW_L3_META_IPV4;
	result = doca_flow_pipe_cfg_set_match(pipe_cfg, &match, NULL);
	if (result != DOCA_SUCCESS) {
		DOCA_LOG_ERR("Failed to set doca_flow_pipe_cfg match: %s", doca_error_get_descr(result));
		goto destroy_pipe_cfg;
	}

	/* prepare copy action */
	desc.type = DOCA_FLOW_ACTION_COPY;
	desc.field_op.src.field_string = "tunnel.esp.spi";
	desc.field_op.dst.field_string = "meta.data";
	desc.field_op.width = 32;
	result = doca_flow_pipe_cfg_set_actions(pipe_cfg, NULL, NULL, desc_list, NB_ACTIONS_ARR);
	if (result != DOCA_SUCCESS) {
		DOCA_LOG_ERR("Failed to set doca_flow_pipe_cfg descs: %s", doca_error_get_descr(result));
		goto destroy_pipe_cfg;
	}

	counter.counter_type = DOCA_FLOW_RESOURCE_TYPE_NON_SHARED;
	result = doca_flow_pipe_cfg_set_monitor(pipe_cfg, &counter);
	if (result != DOCA_SUCCESS) {
		DOCA_LOG_ERR("Failed to set doca_flow_pipe_cfg counter: %s", doca_error_get_descr(result));
		goto destroy_pipe_cfg;
	}

	/* RSS queue - destribute matched traffic between queues */
	for (i = 0; i < nb_rss_queues; ++i)
		rss_queues[i] = i + 1;

	fwd.type = DOCA_FLOW_FWD_RSS;
	fwd.rss_queues = rss_queues;
	fwd.rss_outer_flags = DOCA_FLOW_RSS_ESP;
	fwd.num_of_queues = nb_rss_queues;

	fwd_miss.type = DOCA_FLOW_FWD_DROP;

	result = doca_flow_pipe_create(pipe_cfg, &fwd, &fwd_miss, pipe);
destroy_pipe_cfg:
	doca_flow_pipe_cfg_destroy(pipe_cfg);
	return result;
}

/*
 * Add DOCA Flow pipe entry.
 *
 * @pipe [in]: pipe of the entry.
 * @port [in]: port of the entry.
 * @entry [out]: created entry pointer.
 * @return: DOCA_SUCCESS on success and DOCA_ERROR otherwise
 */
static doca_error_t add_rss_esp_pipe_entry(struct doca_flow_pipe *pipe,
					   struct doca_flow_port *port,
					   struct doca_flow_pipe_entry **entry)
{
	struct doca_flow_match match;
	struct entries_status status;
	doca_error_t result;

	memset(&match, 0, sizeof(match));
	memset(&status, 0, sizeof(status));

	result = doca_flow_pipe_add_entry(0, pipe, &match, NULL, NULL, NULL, 0, &status, entry);
	if (result != DOCA_SUCCESS) {
		DOCA_LOG_ERR("Failed to insert ESP RSS entry: %s", doca_error_get_descr(result));
		return result;
	}

	result = doca_flow_entries_process(port, 0, DEFAULT_TIMEOUT_US, 1);
	if (result != DOCA_SUCCESS) {
		DOCA_LOG_ERR("Failed to process ESP RSS entry: %s", doca_error_get_descr(result));
		return result;
	}

	if (status.nb_processed != 1 || status.failure) {
		DOCA_LOG_ERR("Failed to process ESP RSS entry, nb_processed=%d", status.nb_processed);
		return DOCA_ERROR_BAD_STATE;
	}

	return DOCA_SUCCESS;
}

/*
 * Dequeue packets from DPDK queues and calculates the distribution.
 *
 * @port_id [in]: port id for dequeue packets.
 * @nb_queues [in]: number of RSS queues.
 * @entry [in]: entry sent packets before distribution.
 * @return: DOCA_SUCCESS on success and DOCA_ERROR otherwise.
 */
static doca_error_t rss_distribution_results(uint16_t port_id, uint16_t nb_queues, struct doca_flow_pipe_entry *entry)
{
	struct rte_mbuf *packets[PACKET_BURST];
	struct doca_flow_query stats;
	double actuall_percentage;
	uint32_t total_packets;
	uint16_t nb_packets;
	uint16_t queue_index;
	doca_error_t result;
	int i, j;

	result = doca_flow_query_entry(entry, &stats);
	if (result != DOCA_SUCCESS) {
		DOCA_LOG_ERR("Port %u failed to query RSS ESP entry: %s", port_id, doca_error_get_descr(result));
		return result;
	}

	total_packets = stats.total_pkts;
	if (total_packets == 0) {
		DOCA_LOG_DBG("Port %d doesn't recieve any packet", port_id);
		return DOCA_SUCCESS;
	}

	DOCA_LOG_INFO("Port %d RSS distribution information:", port_id);

	for (i = 0; i < nb_queues; i++) {
		queue_index = i + 1;
		nb_packets = rte_eth_rx_burst(port_id, queue_index, packets, PACKET_BURST);
		actuall_percentage = GET_PERCENTAGE(nb_packets, total_packets);

		DOCA_LOG_INFO("Queue %u received %u packets which is %g%% of the traffic (%u/%u)",
			      queue_index,
			      nb_packets,
			      actuall_percentage,
			      nb_packets,
			      total_packets);

		for (j = 0; j < nb_packets; j++) {
			uint32_t metadata = 0;

			if (rte_flow_dynf_metadata_avail())
				metadata = *RTE_FLOW_DYNF_METADATA(packets[j]);

			DOCA_LOG_DBG("Queue %u packet %d ESP SPI is 0x%08x", queue_index, j, metadata);
		}
	}

	return DOCA_SUCCESS;
}

/*
 * Run flow_rss_esp sample
 *
 * @nb_queues [in]: number of queues the sample will use
 * @return: DOCA_SUCCESS on success and DOCA_ERROR otherwise
 */
doca_error_t flow_rss_esp(int nb_queues)
{
	const int nb_ports = 2;
	struct flow_resources resource = {.nr_counters = 1};
	uint32_t nr_shared_resources[SHARED_RESOURCE_NUM_VALUES] = {0};
	struct doca_flow_port *ports[nb_ports];
	struct doca_dev *dev_arr[nb_ports];
	struct doca_flow_pipe_entry *entries[nb_ports];
	struct doca_flow_pipe *pipe;
	int nb_rss_queues = nb_queues - 1;
	doca_error_t result;
	int port_id;

	result = init_doca_flow(nb_queues, "vnf,hws", &resource, nr_shared_resources);
	if (result != DOCA_SUCCESS) {
		DOCA_LOG_ERR("Failed to init DOCA Flow: %s", doca_error_get_descr(result));
		return result;
	}

	memset(dev_arr, 0, sizeof(struct doca_dev *) * nb_ports);
	result = init_doca_flow_ports(nb_ports, ports, true, dev_arr);
	if (result != DOCA_SUCCESS) {
		DOCA_LOG_ERR("Failed to init DOCA ports: %s", doca_error_get_descr(result));
		doca_flow_destroy();
		return result;
	}

	for (port_id = 0; port_id < nb_ports; port_id++) {
		result = create_rss_esp_pipe(ports[port_id], nb_rss_queues, &pipe);
		if (result != DOCA_SUCCESS) {
			DOCA_LOG_ERR("Port %d failed to create pipe: %s", port_id, doca_error_get_descr(result));
			stop_doca_flow_ports(nb_ports, ports);
			doca_flow_destroy();
			return result;
		}

		result = add_rss_esp_pipe_entry(pipe, ports[port_id], &entries[port_id]);
		if (result != DOCA_SUCCESS) {
			DOCA_LOG_ERR("Port %d failed to add entry: %s", port_id, doca_error_get_descr(result));
			stop_doca_flow_ports(nb_ports, ports);
			doca_flow_destroy();
			return result;
		}
	}

	DOCA_LOG_INFO("Wait %d seconds for packets to arrive", WAITING_TIME);
	sleep(WAITING_TIME);

	for (port_id = 0; port_id < nb_ports; port_id++) {
		result = rss_distribution_results(port_id, nb_rss_queues, entries[port_id]);
		if (result != DOCA_SUCCESS) {
			DOCA_LOG_ERR("Port %d failed to query entry: %s", port_id, doca_error_get_descr(result));
			stop_doca_flow_ports(nb_ports, ports);
			doca_flow_destroy();
			return result;
		}
	}

	result = stop_doca_flow_ports(nb_ports, ports);
	doca_flow_destroy();
	return result;
}
